#!/bin/python3
#%%
import io, os, re
import openpyxl
from openpyxl.styles import Font
from openpyxl.styles import Alignment
from openpyxl.styles import Color
from openpyxl.drawing.image import Image
from openpyxl.formatting.rule import DataBarRule, ColorScaleRule
from openpyxl.utils import get_column_letter, column_index_from_string
from openpyxl.drawing.spreadsheet_drawing import TwoCellAnchor, OneCellAnchor, AnchorMarker
from openpyxl.utils.units import pixels_to_EMU

import pandas as pd

#%%


def cell_set_bold(cell: openpyxl.cell.cell.Cell):
    cell.font = Font(bold=True)
    return None


def cell_set_number_format(cell: openpyxl.cell.cell.Cell, format = '0.000'):
    cell.number_format = format
    return None


def cell_set_alignment(cell: openpyxl.cell.cell.Cell, horizontal='center', vertical='center'):
    cell.alignment = Alignment(horizontal=horizontal, vertical=vertical)
    return None


def sheet_set_dimensions(worksheet: openpyxl.worksheet.worksheet.Worksheet, column = None, column_width = None, index = None, index_height = None):
    if isinstance(column_width, str) and column_width.endswith('px'):
        column_width = float(column_width[:-2]) * .125
    if isinstance(index_height, str) and index_height.endswith('px'):
        index_height = float(index_height[:-2]) * .75
    if column is not None:
        worksheet.column_dimensions[column].width = column_width
    elif index is not None:
        worksheet.row_dimensions[index].height = index_height
    return None


def resize_img(width, height, max_size, ratio = 1):
    if not isinstance(max_size, list):
        max_size = [max_size] * 2
    max_width, max_height = max_size
    print(f'old: width/height = {width}/{height} ({width/height})')
    new_height = max_width / width * height
    new_width = max_height / height * width
    if new_height <= max_height:
        width, height = max_width, new_height
    elif new_width <= max_width:
        width, height = new_width, max_height
    else:
        raise ValueError
    width *= ratio
    height *= ratio
    print(f'new: width/height = {width}/{height} ({width/height})')
    return width, height


def split_string(s):
    match = re.match(r"([a-zA-Z]+)(\d+)", s)
    if match:
        return match.groups()
    else:
        return None


def column_to_letters(worksheet: openpyxl.worksheet.worksheet.Worksheet, columns = []):
    if isinstance(columns, str):
        column = columns
        column_idx = None
        for col in range(1, worksheet.max_column + 1):
            if worksheet.cell(row=1, column=col).value == column:
                column_idx = col
                break
        column_idx = get_column_letter(column_idx)
    elif isinstance(columns, list):
        column_idx = [column_to_letters(worksheet=worksheet, columns=column) for column in columns]
    return column_idx


def insert_dataframe_to_worksheet(worksheet: openpyxl.worksheet.worksheet.Worksheet, df: pd.DataFrame, index_and_columns = True):

    n_rows, n_cols = df.shape

    if index_and_columns:

        gap = 2

        if df.index.name is None:
            index_name = 'index'
        else:
            index_name = df.index.name
        worksheet.cell(1, 1).value = index_name

        for r in range(n_rows):
            worksheet.cell(r+2, 1).value = df.index.to_series().iloc[r]
            del r
        
        for c in range(n_cols):
            worksheet.cell(1, c+2).value = df.columns.to_series().iloc[c]
            del c
    
    else:
        gap = 1

    for r in range(n_rows):
        for c in range(n_cols):
            worksheet.cell(r+gap, c+gap).value = df.iloc[r, c]
            
    return worksheet


def sheet_to_dataframe(sheet: openpyxl.worksheet.worksheet.Worksheet, index_col = 'auto', header = True):
    data = sheet.values
    if header:
        df = pd.DataFrame(columns=next(data))
    else:
        df = pd.DataFrame(columns=[f'{k}' for k in range(sheet.max_column)])
    k = 0
    while True:
        try:
            df.loc[k] = next(data)
            k += 1
        except StopIteration :
            break
    if index_col == 'auto':
        index_col = 'index'
    if index_col is not None:
        df.set_index(keys=index_col, inplace=True)
    return df


def fix_columns_of_data_rule(columns: list):
    if any([isinstance(k, str) for k in columns]) and any([isinstance(k, list) for k in columns]):
        columns = [[k] if isinstance(k, str) else k for k in columns]
    return columns


def flatten_columns(columns: list):
    if isinstance(columns, str):
        columns = [columns]
    new_columns = []
    if all([isinstance(k, str) for k in columns]):
        new_columns = sorted(set(columns))
        print('columns =', columns)
    else:
        for a in columns:
            if isinstance(a, str):
                new_columns.append(a)
            elif isinstance(a, list):
                new_columns.extend(a)
            else:
                raise ValueError(f'found a = {a} in columns.')
        new_columns = flatten_columns(new_columns)
    return new_columns


#%%

class SheetModifier():

    def __init__(self, worksheet: openpyxl.worksheet.worksheet.Worksheet, row1_as_col = False) -> None:
        self.worksheet = worksheet
        self.row1_as_col = row1_as_col
        return None
    
    def __repr__(self) -> str:
        return str(self.to_frame())
    
    def __str__(self) -> str:
        return str(self.__repr__())
    
    def to_frame(self, index_col = 'auto'):
        worksheet = self.worksheet
        df = sheet_to_dataframe(worksheet, index_col=index_col)
        return df

    def set_bold_sereis(self, column = None, index = None, start = 0):
        worksheet = self.worksheet
        if column is not None:
            for cell in worksheet[column][start:]:
                cell_set_bold(cell=cell)
        if index is not None:
            for cell in worksheet[index][start:]:
                cell_set_bold(cell=cell)
        return None
    
    def set_cols_bold(self, columns = [], start = 0):
        print('set_cols_bold()')
        columns = flatten_columns(columns=columns)
        worksheet = self.worksheet
        if self.row1_as_col:
            columns = column_to_letters(worksheet, columns=columns)
        for column in columns:
            self.set_bold_sereis(column=column, start=start)
        return None

    def set_hidden_sereis(self, column = None, index = None, start = 0):
        worksheet = self.worksheet
        if column is not None:
            worksheet.column_dimensions[column].hidden = True
        if index is not None:
            worksheet.row_dimensions[index].hidden = True
        return None
    
    def set_cols_hidden(self, columns = [], start = 0):
        print('set_cols_hidden()')
        columns = flatten_columns(columns=columns)
        worksheet = self.worksheet
        if self.row1_as_col:
            columns = column_to_letters(worksheet, columns=columns)
        for column in columns:
            self.set_hidden_sereis(column=column, start=start)
        return None

    def set_number_format(self, column = None, index = None, format = '0.000', start = 0):
        worksheet = self.worksheet
        if column is not None:
            for cell in worksheet[column][start:]:
                cell_set_number_format(cell=cell, format=format)
        if index is not None:
            for cell in worksheet[index][start:]:
                cell_set_number_format(cell=cell, format=format)
        return None
    
    def set_cols_number_format(self, columns = [], format = '0.000', start = 0):
        print(f'set_cols_number_format()')
        columns = flatten_columns(columns=columns)
        worksheet = self.worksheet
        if self.row1_as_col:
            columns = column_to_letters(worksheet, columns=columns)
        for column in columns:
            self.set_number_format(column=column, format=format, start=start)
        return None
    
    def set_alignment(self, column = None, index = None, horizontal='center', vertical='center', start = 0):
        worksheet = self.worksheet
        if column is not None:
            for cell in worksheet[column][start:]:
                cell_set_alignment(cell=cell, horizontal = horizontal, vertical=vertical)
        if index is not None:
            for cell in worksheet[index][start:]:
                cell_set_alignment(cell=cell, horizontal = horizontal, vertical=vertical)
        return None
    
    def set_cols_alignment(self, columns = [], horizontal='center', vertical='center', start = 0):
        print('set_cols_alignment()')
        columns = flatten_columns(columns=columns)
        worksheet = self.worksheet
        if self.row1_as_col:
            columns = column_to_letters(worksheet, columns=columns)
        for column in columns:
            self.set_alignment(column=column, horizontal=horizontal, vertical=vertical, start=start)
        return None
    

    def set_dimensions(self, column = None, column_width = None, index = None, index_height = None):
        worksheet = self.worksheet
        sheet_set_dimensions(worksheet=worksheet, column = column, column_width = column_width, index = index, index_height = index_height)
        return None
    
    def set_cols_dimensions(self, columns = [], column_width = None):
        print('set_cols_dimensions()')
        columns = flatten_columns(columns=columns)
        worksheet = self.worksheet
        if self.row1_as_col:
            columns = column_to_letters(worksheet, columns=columns)
        for column in columns:
            self.set_dimensions(column = column, column_width = column_width)
        return None
    
    def set_indexes_dimensions(self, indexes = None, index_height = None, start = 2):
        print('set_indexes_dimensions()')
        worksheet = self.worksheet
        if indexes is None:
            indexes = range(start, worksheet.max_row + 1)
        for index in indexes:
            self.set_dimensions(index = index, index_height = index_height)
        return None
    

    def insert_image(self, anchor: str, img = None, resize = 'auto', func = None, ratio = .95):

        worksheet = self.worksheet

        if isinstance(img, str) and os.path.isfile(img):
            with open(img, 'rb') as f:
                img = f.read()
        elif img is None and func is not None:
            cell_value = worksheet[anchor].value
            img = func(cell_value)
        if isinstance(img, bytes):
            img = io.BytesIO(img)
            img = Image(img=img)
        
        if resize == 'auto':
            col, row = split_string(anchor)
            row = int(row)
            print(f'anchor = {anchor} -> {(row, col)}')
            col_num = column_index_from_string(col)
            cell_width = worksheet.column_dimensions[col].width
            cell_height = worksheet.row_dimensions[row].height
            if cell_width is None:
                cell_width = 8.38
            if cell_height is None:
                cell_height = 13.5
            cell_width_pix = cell_width * 8 
            cell_height_pix = cell_height * 4/3 
            print(f'size({(row, col)}) = {(cell_width_pix, cell_height_pix)}')
            img_width, img_height = img.width, img.height
            img_width, img_height = resize_img(img_width, img_height, max_size=[cell_width_pix, cell_height_pix], ratio = ratio)
            
            col_off = pixels_to_EMU((cell_width_pix - img_width) / 2)
            row_off = pixels_to_EMU((cell_height_pix - img_height) / 2)
            _from = AnchorMarker(col_num-1, col_off, row-1, row_off) # dst
            to = AnchorMarker(col_num, -col_off, row, -row_off)
            # cell_width_emu = pixels_to_EMU(cell_width_pix)
            # cell_height_emu = pixels_to_EMU(cell_height_pix)
            # to = AnchorMarker(
            #     col_num-1, cell_width_emu-col_off, 
            #     row-1, cell_height_emu-row_off
            #     )
            # anchor = OneCellAnchor(_from = to)
            # img.anchor = TwoCellAnchor('twoCell', _from, to)
            anchor = TwoCellAnchor('twoCell', _from, to)

        # worksheet.add_image(img=img)
        worksheet.add_image(img=img, anchor=anchor)

        return None


    def insert_col_img(self, column: str, img = None, resize = 'auto', func = None, start = 2, ratio = .95):
        print('insert_col_img()')
        worksheet = self.worksheet
        if self.row1_as_col:
            column = column_to_letters(worksheet, columns=column)
        for k in range(start, worksheet.max_row+1):
            self.insert_image(anchor = f'{column}{k}', img = img, resize = resize, func = func, ratio=ratio)
        return None
    

    def add_data_rule(self, column: str, value_range = 'min-max', rule = 'ColorScale', start = 2):
        
        worksheet = self.worksheet

        if rule == 'DataBar':
            rule = DataBarRule(
                start_type='num', start_value=min(value_range), 
                end_type='num', end_value=max(value_range), 
                color=Color('FF638EC6')
                )
        elif rule == 'ColorScale':
            rule = ColorScaleRule(
                start_value=min(value_range),
                start_type='num',
                # start_type='min',
                start_color=Color(rgb='63BE7B'),  # green
                mid_value=value_range[1] if len(value_range) == 3 else (min(value_range)+max(value_range))/2,
                mid_type='num',
                # mid_type='percentile',
                mid_color=Color(rgb='FFEB84'), # yellow
                end_value=max(value_range),
                end_type='num',
                # end_type='max',
                end_color=Color(rgb='F8696B')    # red
                )
        n_rows = worksheet.max_row
        worksheet.conditional_formatting.add(f'{column}{start}:{column}{n_rows}', rule)

        return None
    
    
    def add_cols_data_rule(self, columns: str, value_range = 'min-max', rule = 'ColorScale', start = 2):
        columns = fix_columns_of_data_rule(columns=columns)
        flag = all([isinstance(k, str) for k in columns])
        if flag:
            print('add_cols_data_rule()')
            worksheet = self.worksheet
            if self.row1_as_col:
                sheet_df = sheet_to_dataframe(worksheet)
                col_df = sheet_df[columns]
                col_min = col_df.min().min()
                col_max = col_df.max().max()
                if isinstance(value_range, str) and value_range == 'min-max':
                    value_range = [col_min, col_max]
                    print(f'columns = {columns}, value_range = \'min-max\' -> {value_range}')

                columns = column_to_letters(worksheet, columns=columns)
            for column in columns:
                self.add_data_rule(column = column, value_range = value_range, rule = rule, start = start)
        else:
            for column in columns:
                self.add_cols_data_rule(columns=column, value_range=value_range, rule=rule, start=start)
        return None
    

    def set_filter(self, freeze_panes = 'A2'):
        worksheet = self.worksheet
        max_row = worksheet.max_row
        max_column_letter = get_column_letter(worksheet.max_column)
        worksheet.auto_filter.ref = f"A1:{max_column_letter}{max_row}"
        worksheet.freeze_panes = freeze_panes
        return None
        

#%%

class ExcelModifier():

    def __init__(self, xlsx_path = None, workbook = None, df = None, dfs = None, row1_as_col = False) -> None:

        if isinstance(workbook, openpyxl.Workbook):
            pass

        elif isinstance(xlsx_path, str):
            workbook = openpyxl.load_workbook(xlsx_path)

        elif isinstance(dfs, list) or isinstance(dfs, dict) or isinstance(df, pd.DataFrame):
            if isinstance(df, pd.DataFrame):
                dfs = {'Sheet1': df}
            if isinstance(dfs, list):
                dfs = {f'Sheet{k}': v for k, v in enumerate(dfs, start=0)}
            workbook = openpyxl.Workbook()
            del workbook['Sheet']
            for sheet_name, df in dfs.items():
                worksheet = workbook.create_sheet(sheet_name)
                insert_dataframe_to_worksheet(worksheet=worksheet, df=df)
        
        self.row1_as_col = row1_as_col
        self.workbook = workbook
        self.sheetnames = self.workbook.sheetnames

        print('ExcelModifier()')
        print(f'self.row1_as_col = {row1_as_col}')
        print(f'self.workbook = {workbook}')
        print(f'self.sheetnames = {self.workbook.sheetnames}')

        return None
    

    def __repr__(self) -> str:
        string = f'sheetnames = {self.sheetnames}'
        return string
    

    def __str__(self) -> str:
        return self.__repr__()
    
    
    def __getitem__(self, sheet_name):
        if isinstance(sheet_name, str):
            ret = SheetModifier(
                self.workbook[sheet_name], 
                row1_as_col=self.row1_as_col)
        elif isinstance(sheet_name, slice):
            ret = self
        return ret
    

    def set_sheets_bolds(self, sheetnames = None, **kwargs):
        print('set_sheets_bolds()')
        if sheetnames is None:
            sheetnames = self.sheetnames
        for sheet_name in sheetnames:
            self[sheet_name].set_bold_sereis(**kwargs)
        return None
    
    
    def set_sheets_alignment(self, sheetnames = None, **kwargs):
        print('set_sheets_alignment()')
        if sheetnames is None:
            sheetnames = self.sheetnames
        for sheet_name in sheetnames:
            self[sheet_name].set_alignment(**kwargs)
        return None
    

    def set_sheets_filter(self, sheet_names = None, **kwargs):
        """
        save()

        :param freeze_panes: freeze_panes.
        :type param1: str

        :return: return None.
        :rtype: NoneType

        :raises ValueError: ValueError
        """
        print('set_sheets_filter()')
        if sheet_names is None:
            sheet_names = self.sheetnames
        for sheet_name in sheet_names:
            self[sheet_name].set_filter(**kwargs)
        return None
    

    def save(self, filename):
        """
        save()

        :param filename: path to file.
        :type param1: str

        :return: return None.
        :rtype: NoneType

        :raises ValueError: ValueError
        """
        print(f'save(filename = \'{filename}\')')
        self.workbook.save(filename=filename)
        return None
    



#%%

    # def set_bold_sereis(self, sheet_name: str, column = None, index = None, start = 0):
    #     worksheet = self.workbook[sheet_name]
    #     if column is not None:
    #         for cell in worksheet[column][start:]:
    #             cell_set_bold(cell=cell)
    #     if index is not None:
    #         for cell in worksheet[index][start:]:
    #             cell_set_bold(cell=cell)
    #     return None
    
    # def set_cols_bold(self, sheet_name: str, columns = [], start = 0):
    #     if self.row1_as_col:
    #         columns = column_to_letters(self.workbook[sheet_name], columns=columns)
    #     for column in columns:
    #         self.set_bold_sereis(sheet_name=sheet_name, column=column, start=start)
    #     return None

    # def set_number_format(self, sheet_name: str, column = None, index = None, format = '0.000', start = 0):
    #     worksheet = self.workbook[sheet_name]
    #     if column is not None:
    #         for cell in worksheet[column][start:]:
    #             cell_set_number_format(cell=cell, format=format)
    #     if index is not None:
    #         for cell in worksheet[index][start:]:
    #             cell_set_number_format(cell=cell, format=format)
    #     return None
    
    # def set_cols_number_format(self, sheet_name: str, columns = [], format = '0.000', start = 0):
        
    #     if self.row1_as_col:
    #         columns = column_to_letters(self.workbook[sheet_name], columns=columns)
    #     for column in columns:
    #         self.set_number_format(sheet_name=sheet_name, column=column, format=format, start=start)
    #     return None

    # def set_alignment(self, sheet_name: str, column = None, index = None, horizontal='center', vertical='center', start = 0):
    #     worksheet = self.workbook[sheet_name]
    #     if column is not None:
    #         for cell in worksheet[column][start:]:
    #             cell_set_alignment(cell=cell, horizontal = horizontal, vertical=vertical)
    #     if index is not None:
    #         for cell in worksheet[index][start:]:
    #             cell_set_alignment(cell=cell, horizontal = horizontal, vertical=vertical)
    #     return None
    
    # def set_cols_alignment(self, sheet_name: str, columns = [], horizontal='center', vertical='center', start = 0):
    #     if self.row1_as_col:
    #         columns = column_to_letters(self.workbook[sheet_name], columns=columns)
    #     for column in columns:
    #         self.set_alignment(sheet_name=sheet_name, column=column, horizontal=horizontal, vertical=vertical, start=start)
    #     return None
    
    # def set_dimensions(self, sheet_name: str, column = None, column_width = None, index = None, index_height = None):
    #     worksheet = self.workbook[sheet_name]
    #     sheet_set_dimensions(worksheet=worksheet, column = column, column_width = column_width, index = index, index_height = index_height)
    #     return None
    
    # def set_cols_dimensions(self, sheet_name: str, columns = [], column_width = None):
    #     if self.row1_as_col:
    #         columns = column_to_letters(self.workbook[sheet_name], columns=columns)
    #     for column in columns:
    #         self.set_dimensions(sheet_name=sheet_name, column = column, column_width = column_width)
    #     return None
    
    # def set_indexes_dimensions(self, sheet_name: str, indexes = None, index_height = None, start = 2):
    #     if indexes is None:
    #         indexes = range(start, self.workbook[sheet_name].max_row + 1)
    #     for index in indexes:
    #         self.set_dimensions(sheet_name=sheet_name, index = index, index_height = index_height)
    #     return None
    
    # def add_data_bar(self, sheet_name: str, column: str, value_range = [0, 1], start = 2):
    #     worksheet = self.workbook[sheet_name]
    #     start_value, end_value = value_range
    #     data_bar_rule = DataBarRule(
    #         start_type='num', start_value=start_value, 
    #         end_type='num', end_value=end_value, 
    #         color=Color('FF638EC6')
    #         )
    #     n_rows = worksheet.max_row
    #     worksheet.conditional_formatting.add(f'{column}{start}:{column}{n_rows}', data_bar_rule)
    #     return None
    
    # def add_cols_data_bar(self, sheet_name: str, columns: str, value_range = [0, 1], start = 2):
    #     if self.row1_as_col:
    #         columns = column_to_letters(self.workbook[sheet_name], columns=columns)
    #     for column in columns:
    #         self.add_data_bar(sheet_name=sheet_name, column = column, value_range = value_range, start = start)
    #     return None
    
    # def insert_image(self, sheet_name: str, anchor: str, img = None, resize = 'auto', func = None):

    #     worksheet = self.workbook[sheet_name]

    #     if isinstance(img, str) and os.path.isfile(img):
    #         with open(img, 'rb') as f:
    #             img = f.read()
    #     elif img is None and func is not None:
    #         cell_value = worksheet[anchor].value
    #         img = func(cell_value)
    #     if isinstance(img, bytes):
    #         img = io.BytesIO(img)
    #         img = Image(img=img)
        
    #     if resize == 'auto':
    #         col, row = split_string(anchor)
    #         print((row, col))
    #         col_num = column_index_from_string(col)
    #         row = int(row)
    #         cell_width = worksheet.column_dimensions[col].width
    #         cell_height = worksheet.row_dimensions[row].height
    #         if cell_width is None:
    #             cell_width = 8.38
    #         if cell_height is None:
    #             cell_height = 13.5
    #         cell_width_pix = cell_width * 8 
    #         cell_height_pix = cell_height * 4/3 
    #         print((row, col), (cell_width_pix, cell_height_pix))
    #         img_width, img_height = img.width, img.height
    #         img_width, img_height = resize_img(img_width, img_height, max_size=[cell_width_pix, cell_height_pix], ratio = .95)
            
    #         col_off = pixels_to_EMU((cell_width_pix - img_width) / 2)
    #         row_off = pixels_to_EMU((cell_height_pix - img_height) / 2)
    #         _from = AnchorMarker(col_num-1, col_off, row-1, row_off)
    #         to = AnchorMarker(col_num, -col_off, row, -row_off)
    #         # img.anchor = TwoCellAnchor('twoCell', _from, to)
    #         anchor = TwoCellAnchor('twoCell', _from, to)

    #     # worksheet.add_image(img=img)
    #     worksheet.add_image(img=img, anchor=anchor)
        
    #     return None
    
    
    # def insert_col_img(self, sheet_name: str, column: str, img = None, resize = 'auto', func = None, start = 2):
    #     worksheet = self.workbook[sheet_name]
    #     if self.row1_as_col:
    #         column = column_to_letters(worksheet, columns=column)
    #     for k in range(start, worksheet.max_row+1):
    #         self.insert_image(sheet_name, anchor = f'{column}{k}', img = img, resize = resize, func = func)
    
    # def set_filter(self, sheet_name: str):
    #     worksheet = self.workbook[sheet_name]
    #     max_row = worksheet.max_row
    #     max_column_letter = get_column_letter(worksheet.max_column)
    #     worksheet.auto_filter.ref = f"A1:{max_column_letter}{max_row}"
    #     return None
    
